#include <windows.h>
#include <ocidl.h>   //  IPicture
#include <olectl.h>  //  OleLoadPicture

#include "jpegmarker.h"
#include "jpeg_read.h"

//  comment this line out to draw black window background 
// #define USE_SYS_BG_COLOR  1

typedef enum {       //Exif Data type
   ED_UB    = 1,     //unsigned byte
   ED_C     = 2,     //ascii string
   ED_US    = 3,     //unsigned short
   ED_UL    = 4,     //unsigned long
   ED_UR    = 5,     //unsigned rational
   ED_B     = 6,     //signed byte
   ED_UD    = 7,     //undefined
   ED_S     = 8,     //signed short
   ED_L     = 9,     //signed long
   ED_R     = 10,    //signed rational
   ED_F     = 11,    //single float
   ED_D     = 12     //double float
};

//***********************************************************************
typedef struct {
   HBITMAP hBm;
   union {
      DWORD size;
      struct {WORD width, height; };
   };
   WORD no;       // file name number
   BYTE orient;
   BYTE type;
} t_EIMAGE;

static t_EIMAGE m_image;

static boolean intel = TRUE;
//***********************************************************************
static WORD GetWord(PBYTE pByte)
{
   if (intel) return *PWORD(pByte);
   else return pByte[0] * 256 + pByte[1];
}

static void SetWord(PBYTE pByte, WORD data)
{
   if (intel) *PWORD(pByte) = data;
   else pByte[0] = BYTE(data >> 8), pByte[1] = BYTE(data);
}

static UINT GetUint(PBYTE pByte)
{
   if (intel) return *((UINT*)(pByte));
   else return UINT(GetWord(pByte)<< 16) + GetWord(pByte+2);
}

// Compare a marker with a list of markers

static bool compare(BYTE b, const BYTE markers[])
{
   for (BYTE n = 1; n <= markers[0]; n++)
      if (markers[n] == b) return true;
   return false;
}
// Search a list of markers in a memory buffer

static bool SearchMarker(PBYTE &pByte, UINT &length, // memory buffer
                         const BYTE markers[])     // list of markers
{
   PBYTE pCur = pByte, pEnd = pByte + length;
   if (pCur && *pCur++ == 0xFF && *pCur++ == M_SOI) {
      // we are in a jpeg memory buffer
      while (pCur + 6 <= pEnd && *pCur++ == 0xFF) {
         BYTE marker = *pCur++;
         if (marker == M_SOS || marker == M_EOI) break;//end of jpeg

         WORD len = pCur[0] * 256 + pCur[1];
         if (compare(marker, markers)) {
            length =  len - 2;
            pByte = pCur + 2;
            return true;
         }
         pCur += len;
      }
   }
   return false;
}

bool GetExifBuffer(PBYTE &pByte, UINT &len)
{
   if (pByte && len > 30 && strcmp((char*)pByte, "Exif") == 0) {
      switch (pByte[6]) {
      case 'I' : intel = TRUE;  pByte += 6, len -= 6; return true;
      case 'M' : intel = FALSE; pByte += 6, len -= 6; return true;
      }
   }
   return false;
}

///////////////////////////////////////////////////////////////////////////////
// EXIF image orientation (read/rewrite)
static const BYTE marquerAPP1[2] = {1, M_APP1};
static const BYTE markersSOx[14] = {13,M_SOF0,M_SOF1,M_SOF2,M_SOF3,M_SOF5,
            M_SOF6,M_SOF7,M_SOF9,M_SOF10,M_SOF11,M_SOF13,M_SOF14,M_SOF15};
static DWORD GetDWSize(PBYTE pCur)
   { return (pCur[3] << 8) | pCur[4] | (pCur[1] << 24) | (pCur[2] << 16); }

///////////////////////////////////////////////////////////////////////////////
WORD GetOrientation(PBYTE pBuf, UINT len, // input buffer
                    WORD orientation)     // orientation flags to write
{
   if (*pBuf == 0xFF) SearchMarker(pBuf, len, marquerAPP1);
   if ( GetExifBuffer(pBuf, len) ) {

      pBuf += GetUint(pBuf + 4);
      // Search the Orientation of image tag on the IDF0
      WORD count = GetWord(pBuf); pBuf += 2;
      for (WORD n = 0; n < count; n++, pBuf += 12) {
         WORD tag = GetWord(pBuf);
         UINT cnt = GetUint(pBuf + 4);
         if ( tag == 0x0112 && cnt == 1 ) {
            WORD tmp;

            // Read the orientation value
            switch (GetWord(pBuf + 2)) {
            case ED_UB:    // byte
            case ED_B: tmp = pBuf[8]; break;
            case ED_US:    // short
            case ED_S: tmp = GetWord(pBuf + 8); break;
            case ED_UL:    // int
            case ED_L: tmp = WORD(GetUint(pBuf + 8)); break;
            default :  tmp = 0;
            }
            if (orientation != 0) {
               // orientation modified
               SetWord(pBuf + 2, ED_US);        // type
               SetWord(pBuf + 8, orientation);  // value
            }
            return tmp;
         }
      }
   }
   return 0;
}
///////////////////////////////////////////////////////////////////////////////
// Extract Bitmap of the data

static BOOL GetImage(HGLOBAL hGlobal, UINT len, t_EIMAGE *pEImage)
{
   BOOL ok = FALSE;
   static IPicture *pPict = NULL;
   static IStream *pIStream = NULL;
   PBYTE pCur = (PBYTE) GlobalLock( hGlobal );
   WORD orient = GetOrientation(pCur, len, 0);

   if (orient != 0) pEImage->orient = (BYTE)orient;
   SearchMarker(pCur, len, markersSOx);
   pEImage->size = GetDWSize(pCur);
   GlobalUnlock( hGlobal );
   // Create IPicture with the file datas
   if (CreateStreamOnHGlobal( hGlobal, TRUE, &pIStream ) == S_OK ) {
      OleLoadPicture(pIStream, len, FALSE, IID_IPicture, (void**)&pPict);
      if (pPict) {
         pEImage->type = 6;
         pPict->get_Handle((OLE_HANDLE*)&pEImage->hBm);
         ok = TRUE;
      }
//      pIStream->Release();
   }
   GlobalFree(hGlobal);
   return ok;
}

//***********************************************************************
// BOOL GetImagef(CFile &file, t_EIMAGE *pEImage)
BOOL GetImagef(LPTSTR path, t_EIMAGE *pEImage)
{
   BOOL ok = FALSE;
   DWORD FileSizeHigh ;
   HANDLE hFile = CreateFile(path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
                              FILE_ATTRIBUTE_NORMAL, NULL) ;
   if (hFile == INVALID_HANDLE_VALUE) {
      wsprintf(tempstr, "CreateFile: %s", get_system_message()) ;
      MessageBox(NULL, tempstr, "CreateFile", MB_OK) ;
   }
   // HANDLE hTemplateFile          // handle to file with attributes to copy
   // UINT len = (DWORD)file.GetLength();
   DWORD len = GetFileSize(hFile,  &FileSizeHigh) ;
   HGLOBAL hGlobal = GlobalAlloc( GMEM_MOVEABLE|GMEM_SHARE, len );
   memset(pEImage, 0, sizeof(t_EIMAGE));
   if (hGlobal) {
      DWORD rbytes ;
      // file.Read( GlobalLock( hGlobal ), len );
      ReadFile(hFile, GlobalLock( hGlobal ), len, &rbytes, NULL );
      ok = GetImage(hGlobal, len, pEImage);
   }
   return ok;
}

//***********************************************************************
HANDLE MakeImage(LPTSTR path)
{
   // CFile file(path, CFile::modeRead);
   if (m_image.hBm) 
      DeleteObject(m_image.hBm);
   m_image.hBm = NULL;

   // GetImagef(file, &m_image);
   BOOL ok = GetImagef(path, &m_image);
   return (ok) ? m_image.hBm : NULL ;

   // SetScrollSizes(MM_TEXT, CSize(10,10));
}

