//*****************************************************************
//  copy rainbow code from gstuff application
//  Try separate thread for the rainbow drawing function
//                                           
//  Written by:   Daniel D. Miller           
//                                           
//  Last Update:  03/10/03 11:34             
//                                           
//  compile with: g++ -Wall -O3 -s rainbow.cpp -o rainbow.exe -lgdi32
//*****************************************************************
//  RAINBOWV.CPP:  A rainbow simulator.                            
//  From "Astronomical Computing"                                  
//  Sky & Telescope Magazine, February 1991                        
//  Original program was in BASIC and EGA 4-color video mode.      
//                                                                 
//  05/27/98
//    Converted to C and VGA/16-color by Dan Miller.               
//  08/09/02 16:11
//    Converted into Win32 graphics program for more colors
//*****************************************************************


static char szAppName[] = "rainbow" ;
static char szAppDesc[] = "draw simulated rainbow" ;

#include <windows.h>
#include <stdlib.h>
#include <math.h>
#include <process.h>

#include "rainbow.h"

//  comment this line out to sleep on message queue
// #define CONTINUOUS_REDRAW 1

//  comment this line out to draw black window background 
// #define USE_SYS_BG_COLOR  1

//***********************************************************************
LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM) ;

int cxClient = 0 ;
int cyClient = 0 ;

HINSTANCE g_hinst = 0 ;
unsigned xbase, xdiff, ybase, ydiff ;

static void display_current_operation(HWND hwnd);

HWND hwnd ;
//***********************************************************************
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                  PSTR szCmdLine, int iCmdShow)
   {
   MSG         msg ;
   WNDCLASSEX  wndclass ;

   g_hinst = hInstance ;

   wndclass.cbSize        = sizeof (wndclass) ;
   wndclass.style         = CS_HREDRAW | CS_VREDRAW ;
   wndclass.lpfnWndProc   = WndProc ;
   wndclass.cbClsExtra    = 0 ;
   wndclass.cbWndExtra    = 0 ;
   wndclass.hInstance     = hInstance ;
   // wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION) ;
   // wndclass.hIconSm       = LoadIcon (NULL, IDI_APPLICATION) ;
   wndclass.hIcon         = LoadIcon (g_hinst, MAKEINTRESOURCE(RBICON));
   wndclass.hIconSm       = LoadIcon (g_hinst, MAKEINTRESOURCE(RBICON));
   wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW) ;
#ifdef   USE_SYS_BG_COLOR  
   wndclass.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1) ;
#else
   wndclass.hbrBackground = (HBRUSH) GetStockObject (BLACK_BRUSH) ;
#endif
   wndclass.lpszMenuName  = NULL ;
   wndclass.lpszClassName = szAppName ;

   RegisterClassEx (&wndclass) ;

   hwnd = CreateWindow (szAppName, szAppDesc,
                        WS_OVERLAPPEDWINDOW,
                        CW_USEDEFAULT, CW_USEDEFAULT,
                        CW_USEDEFAULT, CW_USEDEFAULT,
                        NULL, NULL, hInstance, NULL) ;

   ShowWindow (hwnd, iCmdShow) ;
   UpdateWindow (hwnd) ;

#ifdef CONTINUOUS_REDRAW
   //  this loop repeatedly redraws the current screen,
   //  any time there is no message on the queue.
   //  It utilizes 100% of CPU time!!
   //  If continuous redraw is not required,
   //  use the GetMessage() loop below, instead.
   while (TRUE) {
      if (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE)) {
         if (msg.message == WM_QUIT)
              break ;
   
         TranslateMessage (&msg) ;
         DispatchMessage (&msg) ;
      }
      else {
         display_current_operation(hwnd) ;
      }
   }
#else
   while (GetMessage (&msg, NULL, 0, 0)) {
      TranslateMessage (&msg) ;
      DispatchMessage (&msg) ;
   }
#endif

   return msg.wParam ;
}

//***********************************************************************
static int keep_running = 0 ;
static DWORD WINAPI StartThread(LPVOID iValue);

LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
   // HDC    hdc ;
   static HANDLE hThread1 ;
   DWORD dwGenericThread;  //  does this need to be global??
   char msgstr[81] ;
   // PAINTSTRUCT ps;

   switch (iMsg) {
    
   case WM_CREATE:
      cxClient = 0 ;
      cyClient = 0 ;
      // _beginthread(Thread, 0, NULL) ;
      keep_running = 1 ;
      hThread1 = CreateThread(NULL, 0, StartThread, NULL, 0, &dwGenericThread);
      if(hThread1 == NULL) {
         DWORD dwError = GetLastError();
         wsprintf(msgstr, "SCM:Error %u in Creating thread\n", (unsigned) dwError) ;
         OutputDebugString(msgstr) ;
         return 1;
      }
      return 0 ;

   case WM_SIZE:
      cxClient = LOWORD (lParam) ;
      cyClient = HIWORD (lParam) ;

      xbase = cxClient / 2 ;     
      xdiff = (cxClient-50) / 2 ;
      ybase = cyClient * 7 / 8 ; 
      ydiff = cyClient * 5 / 8 ; 

      return 0 ;

   case WM_DESTROY:
      keep_running = 0 ;
      WaitForSingleObject(hThread1,INFINITE);
      PostQuitMessage (0) ;
      return 0 ;
   }

   return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}

//**************************************************************************
//  generate random number between 0 and n-1
//**************************************************************************
// In Numerical Recipes in C: The Art of Scientific Computing (William H. 
// Press, Brian P. Flannery, Saul A. Teukolsky, William T. Vetterling;
// New York: Cambridge University Press, 1990 (1st ed, p. 207)),      
// the following comments are made:                                   
//                                                                    
// "If you want to generate a random integer between 1 and 10,        
// you should always do it by                                         
//                                                                    
//    j=1+(int) (10.0*rand()/(RAND_MAX+1.0));                         
//                                                                    
// and never by anything resembling                                   
//                                                                    
//    j=1+((int) (1000000.0*rand()) % 10);                            
//                                                                    
// (which uses lower-order bits)."                                    
//**************************************************************************
int random_int(int n)
{
   //  turn random number into a floating-point percentage
   double d = (double) rand() / (RAND_MAX + 1.0) ;
   //  then calculate result as that random percentage of the target value
   return (int) (n * d) ;
}

//**************************************************************************
//  turn random number into a floating-point percentage
//**************************************************************************
double random_part(void)
{
   return (double) rand() / (RAND_MAX + 1.0) ;
}

//*****************************************************************
//  RAINBOWV.CPP:  A rainbow simulator.                            
//*****************************************************************

unsigned rainbow_index[6] = { 
   0x000000AA,       //  red     
   RGB(255,165,0),   //  orange  
   0x0055FFFF,       //  yellow  
   0x0000AA00,       //  green   
   0x00AA0000,       //  blue    
   0x00AA00AA        //  violet  
} ;
const double RADS2DEGS = (180.0 / 3.14159) ;
double X, Y, B ;

/************************************************************************/
void rainbow_plot_pixel(HDC hdc, int pcolor, double TH)
{
   double f1 ;
   unsigned XP, YP ;

   TH = fabs(TH) ;
   if (TH > 60.0)
      return ;
   f1 = TH / 60.0 ;

   XP = (unsigned) (xbase + xdiff * (f1 *     (X / B))) ;
   YP = (unsigned) (ybase - ydiff * (f1 * fabs(Y / B))) ;
   SetPixel(hdc, XP, YP, rainbow_index[pcolor]) ;
}

/************************************************************************/
//  this function plots one primary and one secondary pixel
/************************************************************************/
void display_current_operation(HWND hwnd)
// void DrawRainbow(HWND hwnd)
{
   HDC hdc ;
   int pcolor ;
   double N, I, R, T1, T2, RS, RP, RB, RC, I1, I2 ;

   if (cxClient == 0 || cyClient == 0)
        return ;

   // sprintf(tempstr, "cx=%u, cy=%u\n", cxClient, cyClient) ;
   // MessageBox(hwnd, tempstr, "status", MB_OK) ;

   // 30 REM  RANDOM IMPACT PARAMETER
   X = -1.0 + 2 * random_part() ;
   Y = -1.0 + 2 * random_part() ;
   B = (double) sqrt(X * X + Y * Y) ; //  square root
   // 50 IF B >= 1 THEN 30
   if (B >= 1.0) {
      return ;
   }

   hdc = GetDC (hwnd) ;
   //  select random color and calculate index of refraction
   pcolor = random_int(6) ; //  select 1 thru 6
   N = 1.33 + 0.01 * (double) (pcolor) ;

   // 70 REM  COMPUTE ANGLES
   I  = asin(B) ;     //  angle of incidence
   R  = asin(B / N) ; //  angle of refraction

   //  calculate observation angles for 
   //  primary (T1) and secondary (T2) rainbows.
   T1 = (4 * R - 2 * I) * RADS2DEGS ;
   T2 = (6 * R - 2 * I) * RADS2DEGS - 180 ;

   // 95 REM  INTENSITY FACTORS
   // RS = (sin(I - R) / sin(I + R)) ^ 2 ;
   RS = (double) (sin(I - R) / sin(I + R)) ;
   RS *= RS ;
   RP = (double) (tan(I - R) / tan(I + R)) ;
   RP *= RP ;
   RB = (1 - RP) * (1 - RP) ;
   RC = (1 - RS) * (1 - RS) ;
   I1 = (     RS * RC +      RP + RB) / 2 ;
   I2 = (RS * RS * RC + RP * RP * RB) / 2 ;

   //  Greenler's method: rather than computing intensities
   //  for individual points, we randomly throw away some
   //  computed points in the low-intensity regions.
   if (I1 >= .04 * random_part())
      rainbow_plot_pixel(hdc, pcolor, T1) ;

   if (I2 >= .02 * random_part())
      rainbow_plot_pixel(hdc, pcolor, T2) ;

   ReleaseDC (hwnd, hdc) ;
}

//*****************************************************************
//  separate thread for drawing function
//*****************************************************************
static DWORD WINAPI StartThread(LPVOID iValue)
{
   while (keep_running) {
      display_current_operation(hwnd);
   }
   return (DWORD) 0;
}

